(* ETH Oberon, Copyright 2000 ETH Zuerich Institut fuer Computersysteme, ETH Zentrum, CH-8092 Zuerich.
Refer to the "General ETH Oberon System Source License" contract available at: http://www.oberon.ethz.ch/ *)

MODULE Fonts; (** portable / source: Win32.Fonts.Mod *)	(* ejz  *)
	IMPORT SYSTEM, Modules, Objects;

	CONST
		substitute* = -1; font* = 0; metric* = 1;	(** Font types. *)

	TYPE 
		Char* = POINTER TO CharDesc;
		Font* = POINTER TO FontDesc;

		CharDesc* = RECORD (Objects.ObjDesc)	(** The objects in a font library. *)
			dx*, x*, y*, w*, h*: INTEGER;	(** Character width, pattern offset (x, y), pattern size (w, h). *)
			pat*: LONGINT	(** Character raster data. *)
		END;

		FontDesc* = RECORD (Objects.LibDesc)
			type*: SHORTINT;	(** Substitute, font, or metric. *)
			height*, minX*, maxX*, minY*, maxY*: INTEGER	(** Font height, extremal values of characters in font. *)
		END;

	VAR
		FontId*: CHAR;	(** Initial character of font files (.Fnt). *)
		Default*: Font;	(** Default system screen font (typically Syntax10.Scn.Fnt). *)

	(** Return the character and data of ch in a font. *)
	PROCEDURE GetChar*(F: Font; ch: CHAR; VAR dx, x, y, w, h: INTEGER; VAR pat: LONGINT);
		VAR obj: Objects.Object;
	BEGIN
		F.GetObj(F, ORD(ch), obj);
		WITH obj: Char DO
			dx := obj.dx; x := obj.x; y := obj.y; w := obj.w; h := obj.h; pat := obj.pat
		END
	END GetChar;

	(** Load and cache a font. *)
	PROCEDURE This*(name: ARRAY OF CHAR): Font;
		VAR L: Objects.Library;
	BEGIN
		L := Objects.ThisLibrary(name);
		IF (L # NIL) & (L IS Font) THEN RETURN L(Font) ELSE RETURN NIL END
	END This;

	PROCEDURE Init();
		VAR mod: Modules.Module;
	BEGIN
		FontId := 0DBX; Default := NIL;
		(*mod := Modules.ThisMod("WinFonts")*)
	END Init;

BEGIN
	Init()
END Fonts.